###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import uuid
import iso8601
import humanize
import pprint
from typing import Optional, Dict, List, Tuple, Any
from pprint import pformat

import six

import zlmdb

import nacl
import nacl.signing
import nacl.encoding

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks, returnValue, DeferredList, Deferred
from twisted.internet.task import LoopingCall
from twisted.web import client

import txaio
from txaio import make_logger, time_ns

from autobahn import wamp
from autobahn.util import utcnow
from autobahn.wamp.types import RegisterOptions, SubscribeOptions, CallDetails, PublishOptions, EventDetails
from autobahn.wamp.exception import ApplicationError
from autobahn.twisted.wamp import ApplicationSession

from crossbar._util import hl, hlid, hlval, hltype
from crossbar.common.key import _read_node_key, _read_release_key

from cfxdb.mrealmschema import MrealmSchema
from cfxdb.globalschema import GlobalSchema

from crossbar.master.api import APIS
from crossbar.master.cluster import WebClusterManager, RouterClusterManager
from crossbar.master.mrealm.metadata import MetadataManager
from crossbar.master.arealm import ApplicationRealmManager

from cfxdb.log import MNodeLog, MWorkerLog

__all__ = ('MrealmController', )

client._HTTP11ClientFactory.noisy = False


class Node(object):
    """
    Run-time representation of CF nodes currently connected to this management realm.
    """

    LAST_ACTIVITY_NONE = 0
    LAST_ACTIVITY_CREATED = 1
    LAST_ACTIVITY_STARTED = 2
    LAST_ACTIVITY_READY = 3
    LAST_ACTIVITY_HEARTBEAT = 4
    LAST_ACTIVITY_CHECK = 5

    def __init__(self,
                 node_id=None,
                 heartbeat_counter=None,
                 heartbeat_time=None,
                 heartbeat_workers={},
                 status='online',
                 node_authid=None):
        self.node_id = node_id
        self.node_authid = node_authid
        self.status = status
        self.last_active = time_ns()
        self.last_activity = Node.LAST_ACTIVITY_CREATED

        self.heartbeat_counter = heartbeat_counter
        self.heartbeat_time = heartbeat_time or 0
        self.heartbeat_workers = heartbeat_workers

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        return {
            'id': self.node_id,
            'authid': self.node_authid,
            'status': self.status,
            'last_active': self.last_active,
            'last_activity': self.last_activity,

            # these data items are received from managed nodes
            'workers': self.heartbeat_workers,
            'counter': self.heartbeat_counter,
            'time': self.heartbeat_time,
        }


class Trace(object):
    def __init__(self, trace_id, traced_workers, trace_options, eligible_reader_roles, exclude_reader_roles, status):
        self.trace_id = trace_id
        self.traced_workers = traced_workers
        self.trace_options = trace_options
        self.eligible_reader_roles = eligible_reader_roles
        self.exclude_reader_roles = exclude_reader_roles
        self.status = status

    def marshal(self):
        return {
            'trace_id': self.trace_id,
            'traced_workers': self.traced_workers,
            'trace_options': self.trace_options,
            'eligible_reader_roles': self.eligible_reader_roles,
            'exclude_reader_roles': self.exclude_reader_roles,
            'status': self.status,
        }


class MrealmController(ApplicationSession):
    """
    Backend of user created management realms.

    When a management realm is created, one instance of this component is started.
    This management realm component is then running continuously during the lifetime
    of the management realm.
    """

    log = make_logger()

    def onUserError(self, fail, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
        """
        if isinstance(fail.value, ApplicationError):
            self.log.debug('{klass}.onUserError(): "{msg}"',
                           klass=self.__class__.__name__,
                           msg=fail.value.error_message())
        else:
            self.log.error(
                '{klass}.onUserError(): "{msg}"\n{traceback}',
                klass=self.__class__.__name__,
                msg=msg,
                traceback=txaio.failure_format_traceback(fail),
            )

    def __init__(self, config=None):
        ApplicationSession.__init__(self, config)

        # comes from "crossbar/master/node/node.py" (L299, L282)
        assert config.extra

        self._uri_prefix = 'crossbarfabriccenter.mrealm.'
        self._nodes = None
        self._node_oid_by_name = None
        self._sessions = None
        self._traces = None

        # background loop run periodically to health check and send heartbeats for this controller
        self._tick_loop = None

        # background loop run periodically to check & apply mrealm-level resources
        self._check_and_apply_loop = None
        self._check_and_apply_in_progress = False

        # Release (public) key
        self._release_pubkey_hex = _read_release_key()['hex']

        # Node key
        self._node_key_hex = _read_node_key('.', private=True)['hex']
        self._node_key = nacl.signing.SigningKey(self._node_key_hex, encoder=nacl.encoding.HexEncoder)

        assert 'mrealm' in config.extra and type(config.extra['mrealm'] == str)
        self._mrealm_oid = uuid.UUID(config.extra['mrealm'])

        # controller database
        #
        dbcfg = config.extra.get('controller-database', {})
        assert dbcfg and type(dbcfg) == dict

        dbfile = dbcfg.get('dbfile', None)
        assert dbfile and type(dbfile) == six.text_type

        maxsize = dbcfg.get('maxsize', None)
        assert maxsize and type(maxsize) in six.integer_types
        assert maxsize >= 2**20 and maxsize < 2**30 * 10  # 1 MB - 10 GB maximum size

        self.gdb = zlmdb.Database(dbpath=dbfile, maxsize=maxsize, readonly=False, sync=True)
        self.gdb.__enter__()
        self.gschema: GlobalSchema = GlobalSchema.attach(self.gdb)

        self.log.info('global database initialized [dbfile={dbfile}, maxsize={maxsize}]',
                      dbfile=hlval(dbfile),
                      maxsize=hlval(maxsize))

        # mrealm database
        #
        dbcfg = config.extra.get('database', {})
        assert dbcfg and type(dbcfg) == dict

        dbfile = dbcfg.get('dbfile', None)
        assert dbfile and type(dbfile) == six.text_type

        maxsize = dbcfg.get('maxsize', None)
        assert maxsize and type(maxsize) in six.integer_types
        assert maxsize >= 2**20 and maxsize < 2**30 * 10  # 1 MB - 10 GB maximum size

        self.db = zlmdb.Database(dbpath=dbfile, maxsize=maxsize, readonly=False, sync=True)
        self.db.__enter__()
        self.schema = MrealmSchema.attach(self.db)

        self.log.info('management realm database initialized [dbfile={dbfile}, maxsize={maxsize}]',
                      dbfile=hlid(dbfile),
                      maxsize=hlval(maxsize))

        self._metadata_manager = MetadataManager(self, self.db, self.schema)
        self._webcluster_manager = WebClusterManager(self, self.gdb, self.gschema, self.db, self.schema)
        self._routercluster_manager = RouterClusterManager(self, self.gdb, self.gschema, self.db, self.schema)
        self._arealm_manager = ApplicationRealmManager(self, self.gdb, self.gschema, self.db, self.schema)

    @property
    def nodes(self):
        """
        Returns handle to map of currently connected nodes. The map is indexed by ``node_id``.

        :return: Map of node object ID to :class:`crossbar.master.mrealm.controller.Node`
        :rtype: dict
        """
        return self._nodes

    def node(self, node_authid):
        """
        Get node by node authid (rather than oid).

        :param node_authid:
        :return:
        """
        with self.gdb.begin() as txn:
            node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]

        # currently, nodes are indexed by str-type UUID in the run-time map
        node_oid = str(node_oid)

        return self._nodes.get(node_oid, None)

    def map_node_oid_to_authid(self, node_oid):
        """
        Map node object ID to node authid. This referes to any node for which there is a run-time representation
        currently active, regardless of whether this node is currently online or not.

        :param node_oid:
        :return:
        """
        node_oid = uuid.UUID(node_oid)
        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, node_oid]
        if node:
            self.log.debug('{func}: mapped node object ID {node_oid} to node authid "{node_authid}"',
                           func=hltype(self.map_node_oid_to_authid),
                           node_oid=hlid(node_oid),
                           node_authid=hlid(node.authid))
            return node.authid
        else:
            self.log.warn('{func}: could not map node object ID {node_oid} to any node authid',
                          func=hltype(self.map_node_oid_to_authid),
                          node_oid=hlid(node_oid))
            return None

    @inlineCallbacks
    def onJoin(self, details):
        # initialize this mrealm run-time representation
        try:
            yield self._initialize(details)
        except:
            # immediately close down if there is any error during mrealm initialization
            self.log.failure()
            self.leave()
        else:
            self.log.info('Management controller started for management realm "{realm}" {func}',
                          func=hltype(self.onJoin),
                          realm=hlid(details.realm))

        # initialize Metadata manager
        yield self._metadata_manager.start(prefix=self._uri_prefix + 'metadata.')

        # initialize Web cluster manager
        yield self._webcluster_manager.start(prefix=self._uri_prefix + 'webcluster.')

        # initialize router cluster manager
        yield self._routercluster_manager.start(prefix=self._uri_prefix + 'routercluster.')

        # initialize application realm manager
        yield self._arealm_manager.start(prefix=self._uri_prefix + 'arealm.')

        self.log.info(
            'Management realm controller ready for management realm {mrealm_oid}! (realm="{realm}", session={session}, authid="{authid}", authrole="{authrole}")  [{func}]',
            mrealm_oid=hlid(self._mrealm_oid),
            realm=hlval(details.realm),
            session=hlval(details.session),
            authid=hlval(details.authid),
            authrole=hlval(details.authrole),
            func=hltype(self.onJoin))

    def onLeave(self, details):
        # first, stop background "check nodes" task
        if self._check_and_apply_loop:
            if self._check_and_apply_loop.running:
                self._check_and_apply_loop.stop()
            self._check_and_apply_loop = None

        # next, stop own heartbeat task
        if self._tick_loop:
            if self._tick_loop.running:
                self._tick_loop.stop()
            self._tick_loop = None

        return ApplicationSession.onLeave(self, details)

    @inlineCallbacks
    def _initialize(self, details):
        # "self.config.extra" comes from "crossbar/master/node/node.py"
        self.log.debug('{klass} starting realm "{realm}" with config:\n{config}',
                       klass=self.__class__.__name__,
                       realm=self._realm,
                       config=pprint.pformat(self.config.extra))

        self._started = utcnow()

        self._nodes = {}
        self._nodes_shutdown = {}
        self._node_oid_by_name = {}
        self._sessions = {}
        self._traces = {}

        # subscribe to node lifecycle events
        yield self.subscribe(self._on_node_ready, 'crossbarfabriccenter.node..on_ready',
                             SubscribeOptions(match='wildcard', details=True))

        # we subscribe to CF node heartbeat events, to track when we have last heard of a specific
        # node, and also to react (after some time) should be neither receive an "on_shutdown" nor
        # an "on_leave" event, so we can purge the node from our active list.
        yield self.subscribe(self._on_node_heartbeat, 'crossbarfabriccenter.node.on_heartbeat',
                             SubscribeOptions(details=True))
        yield self.subscribe(self._on_worker_heartbeat, 'crossbarfabriccenter.node.on_worker_heartbeat',
                             SubscribeOptions(details=True))

        # when a CF node is gracefully shut down, we will receive this event. when the CF node
        # is killed, see "session lifecycle events" below.
        yield self.subscribe(self._on_node_shutdown, 'crossbarfabriccenter.node..on_shutdown',
                             SubscribeOptions(match='wildcard', details=True))

        # subscribe to session lifecycle events.
        yield self.subscribe(self._on_session_startup, 'wamp.session.on_join', SubscribeOptions(details=True))

        # eg when a CF node is hard-killed, the management session will simply get lost, which
        # is detected by CFC router, and a WAMP session leave meta event is published. however,
        # no "on_shutdown" event is published! the CF node has been killed and had no chance to
        # send out any management events. hence we must react to this event.
        yield self.subscribe(self._on_session_shutdown, 'wamp.session.on_leave', SubscribeOptions(details=True))

        # produce CFCs own heartbeat on the management realm
        self._tick = 1
        tick_topic = '{}on_tick'.format(self._uri_prefix)

        @inlineCallbacks
        def on_tick():
            if self.is_attached():
                ticked = {'now': utcnow(), 'tick': self._tick}
                yield self.publish(tick_topic, ticked, options=PublishOptions(acknowledge=True))
            else:
                self.log.warn(
                    'cannot send tick to CF node (topic="{tick_topic}", mrealm="{mrealm}""): management realm controller session no longer attached!',
                    tick_topic=tick_topic,
                    mrealm=self._realm)
            self._tick += 1

        self._tick_loop = LoopingCall(on_tick)
        self._tick_loop.start(5)

        self._check_and_apply_loop = LoopingCall(self.check_and_apply)
        self._check_and_apply_loop.start(10)

        # CFC public API
        #
        # register the public, user facing API of CFC. this is the API that
        # user management components, our own CLI (cbsh) or CFC Web user interface
        # will call into. since this is public API, special care should be taken
        # with the design, keeping the API surface limited, logical and extensible

        # management controller top-level API (on this object)
        regs = yield self.register(self, prefix=self._uri_prefix, options=RegisterOptions(details_arg='details'))
        procs = [reg.procedure for reg in regs]
        self.log.debug('Mrealm controller {api} registered management procedures [{func}]:\n\n{procs}\n',
                       api=hl('Realm Home API', color='green', bold=True),
                       func=hltype(self._initialize),
                       procs=hl(pformat(procs), color='white', bold=True))

        # different remoting APIs for node management
        self._apis = APIS

        # setup procedures of APIs
        all_regs = []
        for api in self._apis:
            regs = api.register(self)
            all_regs.extend(regs)

        regs = yield DeferredList(all_regs)
        procs = [reg.procedure for _, reg in regs]
        self.log.debug('Mrealm controller {api} registered management procedures [{func}]:\n\n{procs}\n',
                       api=hl('Node Remoting API', color='magenta', bold=True),
                       func=hltype(self._initialize),
                       procs=hl(pformat(procs), color='white', bold=True))

        # setup topics of APIs
        all_subs = []
        for api in self._apis:
            subs = api.subscribe(self)
            all_subs.extend(subs)

        subs = yield DeferredList(all_subs)
        topics = [sub.topic for _, sub in subs]
        self.log.debug('Mrealm controller {api} subscribed management topics [{func}]:\n\n{topics}\n',
                       api=hl('Node Remoting API', color='magenta', bold=True),
                       func=hltype(self._initialize),
                       topics=hl(pformat(topics), color='white', bold=True))

        # initialize tracing API
        yield self._init_trace_api()

    @inlineCallbacks
    def check_and_apply(self):
        if self._check_and_apply_in_progress:
            self.log.info('{func} {action} for mrealm {mrealm} skipped! check & apply already in progress.',
                          action=hl('check & apply run skipped', color='red', bold=True),
                          func=hltype(self.check_and_apply),
                          mrealm=hlid(self._mrealm_oid))
            return
        else:
            self.log.info('{func} {action} for mrealm {mrealm} ..',
                          action=hl('check & apply run started', color='green', bold=True),
                          func=hltype(self.check_and_apply),
                          mrealm=hlid(self._mrealm_oid))
            self._check_and_apply_in_progress = True

        if not self.is_attached():
            return

        is_running_completely = True
        cnt_nodes_online = 0
        cnt_nodes_offline = 0
        for node_id in self._nodes:
            try:
                node_status = yield self.call('crossbarfabriccenter.remote.node.get_status', node_id)
            except Exception as e:
                cnt_nodes_offline += 1
                is_running_completely = False

                if isinstance(e, ApplicationError) and e.error == 'wamp.error.no_such_procedure':
                    if self._nodes[node_id].status == 'offline':
                        # this is "expected" - we already knew that the node is offline, and hence the call is failing
                        # because of "no_such_procedure" is exactly what will happen as the node is offline
                        self.log.warn(
                            '{action} [status={status}] {func}',
                            action=hl('Warning, managed node "{}" still not connected or operational'.format(node_id),
                                      color='red',
                                      bold=False),
                            status=hlval(self._nodes[node_id].status),
                            func=hltype(self.check_and_apply))
                    else:
                        if self._nodes[node_id].status == 'online':
                            self._nodes_shutdown[node_id] = time_ns()

                            # mark node as offline in run-time map
                            self._nodes[node_id].status = 'offline'

                            # publish management event
                            yield self._publish_on_node_shutdown_yield(self._nodes[node_id])

                            self.log.info('{action} [oid={node_oid}, session={session_id}, status={status}] {func}',
                                          action=hl('Warning: managed node "{}" became offline'.format(node_id),
                                                    color='red',
                                                    bold=True),
                                          node_oid=hlid(node_id),
                                          session_id=hlid(None),
                                          status=hlval(self._nodes[node_id].status),
                                          func=hltype(self.check_and_apply))

                        else:
                            self.log.warn('{func}: unexpected run-time node status {status} for node {node_id}',
                                          node_id=hlid(node_id),
                                          func=hltype(self.check_and_apply),
                                          status=hlval(self._nodes[node_id].status))
                        self._nodes[node_id].status = 'offline'
                else:
                    self._nodes[node_id].status = 'offline'
                    self.log.warn('{action} [status={status}] {func}',
                                  action=hl('Warning: check on managed node "{}" failed: {}'.format(node_id, e),
                                            color='red',
                                            bold=True),
                                  status=hlval(self._nodes[node_id].status),
                                  func=hltype(self.check_and_apply))
            else:
                cnt_nodes_online += 1
                self._nodes[node_id].heartbeat_workers = node_status['workers_by_type']
                self._nodes[node_id].last_activity = Node.LAST_ACTIVITY_CHECK
                self._nodes[node_id].last_active = time_ns()
                if self._nodes[node_id].status == 'online':
                    self.log.info('{action} [status={status}] {func}',
                                  action=hl('Ok, managed node "{}" is still healthy'.format(node_id),
                                            color='green',
                                            bold=False),
                                  status=hlval(self._nodes[node_id].status),
                                  func=hltype(self.check_and_apply))
                else:
                    self.log.info('{action} [status={status} -> "{new_status}"] {func}',
                                  action=hl('Ok, managed node "{}" became healthy (again)'.format(node_id),
                                            color='yellow',
                                            bold=True),
                                  status=hlval(self._nodes[node_id].status),
                                  new_status=hlval('online'),
                                  func=hltype(self.check_and_apply))
                    self._nodes[node_id].status = 'online'

                    # publish "on_node_ready" management event
                    yield self._publish_on_node_ready_yield(self._nodes[node_id])

        if is_running_completely:
            color = 'green'
            action = 'check & apply run completed successfully'
        else:
            color = 'red'
            action = 'check & apply run finished with problems left'

        self._check_and_apply_in_progress = False
        self.log.info(
            '{func} {action} for mrealm {mrealm}: {cnt_nodes_online} nodes online, {cnt_nodes_offline} nodes offline.',
            action=hl(action, color=color, bold=True),
            func=hltype(self.check_and_apply),
            mrealm=hlid(self._mrealm_oid),
            cnt_nodes_online=hlval(cnt_nodes_online),
            cnt_nodes_offline=hlval(cnt_nodes_offline))

    async def _publish_on_node_ready(self, node):
        options = PublishOptions(acknowledge=True)
        uri = '{}on_node_ready'.format(self._uri_prefix)
        obj = node.marshal()

        await self.publish(uri, node.node_id, obj, options=options)

        await self.config.controller.publish(uri, node.node_id, obj, options=options)
        self.log.debug(
            '.. forward published event to controller for managed node {node_id} (uri={uri}, session_id={session_id}, realm={realm}, authid={authid}, authrole={authrole})',
            node_id=node.node_id,
            uri=uri,
            session_id=self.config.controller._session_id,
            realm=self.config.controller._realm,
            authid=self.config.controller._authid,
            authrole=self.config.controller._authrole,
        )

    @inlineCallbacks
    def _publish_on_node_ready_yield(self, node):
        # FIXME: this is a super hack: we need a twisted thing here in "on_check_nodes". Keep synced to "_publish_on_node_shutdown"!
        options = PublishOptions(acknowledge=True)
        uri = '{}on_node_ready'.format(self._uri_prefix)
        obj = node.marshal()

        yield self.publish(uri, node.node_id, obj, options=options)

        yield self.config.controller.publish(uri, node.node_id, obj, options=options)
        self.log.debug(
            '.. forward published event to controller for managed node {node_id} (uri={uri}, session_id={session_id}, realm={realm}, authid={authid}, authrole={authrole})',
            node_id=node.node_id,
            uri=uri,
            session_id=self.config.controller._session_id,
            realm=self.config.controller._realm,
            authid=self.config.controller._authid,
            authrole=self.config.controller._authrole,
        )

    async def _publish_on_node_shutdown(self, node):
        options = PublishOptions(acknowledge=True)
        uri = '{}on_node_shutdown'.format(self._uri_prefix)
        obj = node.marshal()

        await self.publish(uri, node.node_id, obj, options=options)

        await self.config.controller.publish(uri, node.node_id, obj, options=options)
        self.log.debug(
            '.. forward published event to controller for managed node {node_id} (uri={uri}, session_id={session_id}, realm={realm}, authid={authid}, authrole={authrole})',
            node_id=node.node_id,
            uri=uri,
            session_id=self.config.controller._session_id,
            realm=self.config.controller._realm,
            authid=self.config.controller._authid,
            authrole=self.config.controller._authrole,
        )

    @inlineCallbacks
    def _publish_on_node_shutdown_yield(self, node):
        # FIXME: this is a super hack: we need a twisted thing here in "on_check_nodes". Keep synced to "_publish_on_node_shutdown"!
        options = PublishOptions(acknowledge=True)
        uri = '{}on_node_shutdown'.format(self._uri_prefix)
        obj = node.marshal()

        yield self.publish(uri, node.node_id, obj, options=options)

        yield self.config.controller.publish(uri, node.node_id, obj, options=options)
        self.log.debug(
            '.. forward published event to controller for managed node {node_id} (uri={uri}, session_id={session_id}, realm={realm}, authid={authid}, authrole={authrole})',
            node_id=node.node_id,
            uri=uri,
            session_id=self.config.controller._session_id,
            realm=self.config.controller._realm,
            authid=self.config.controller._authid,
            authrole=self.config.controller._authrole,
        )

    async def _on_node_ready(self, ready_info=None, details: Optional[CallDetails] = None):
        node_id = ready_info.get('node_id', None) if ready_info else None
        self.log.info('Node "{node_id}" is ready: {ready_info} {details}',
                      node_id=node_id,
                      ready_info=ready_info,
                      details=details)

        if node_id in self._nodes:
            self.log.warn('Node-ready event for "{node_id}", but node not already in active node list found!',
                          node_id=node_id)

        _node = Node(node_id)
        _node.last_activity = Node.LAST_ACTIVITY_READY
        _node.last_active = time_ns()

        self._nodes[node_id] = _node

        await self._publish_on_node_ready(_node)

        self.log.debug('{func}: completed!', func=hltype(self._on_node_ready))

    async def _on_worker_heartbeat(self, node_authid, worker_id, heartbeat, details: Optional[CallDetails] = None):
        """
        Receive heartbeats from workers run on CF nodes managed by this CFC instances. By default,
         CF node workers will send one hearbeat every 10 seconds.

        :param node_authid: The node ID (UUID in the master database).
        :type node_authid: str

        :param worker_id: The local worker ID on the remote CF node (_not_ the UUID of the worker
            in the master database)
        :type worker_id: str

        :param heartbeat: Node heartbeat.
        :type heartbeat: dict

        :param details: Event details.
        :type details: :class:`autobahn.wamp.types.EventDetails`
        """
        assert type(node_authid) == str
        assert type(worker_id) == str
        assert type(heartbeat) == dict
        assert details is None or isinstance(details, EventDetails)

        started = time_ns()

        heartbeat_time = heartbeat.get('timestamp', None)
        heartbeat_seq = heartbeat.get('seq', None)
        self.log.debug(
            'Heartbeat from worker "{worker_id}" on node "{node_authid}" [heartbeat_seq={heartbeat_seq}, time={heartbeat_time}, publisher={publisher}, authid={authid}]',
            node_authid=hlid(node_authid),
            worker_id=hlid(worker_id),
            heartbeat_seq=hlid(heartbeat_seq),
            heartbeat_time=hlid(heartbeat_time),
            publisher=hlid(details.publisher) if details else None,
            authid=hlid(details.publisher_authid) if details else None)

        self.log.debug('Raw worker heartbeat: \n{heartbeat}', heartbeat=pprint.pformat(heartbeat))

        with self.gdb.begin() as txn:
            node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]
            node = self.gschema.nodes[txn, node_oid]

        node_oid = str(node_oid)

        mrealm_id = node.mrealm_oid
        mworker_log = MWorkerLog.parse(mrealm_id, uuid.UUID(node_oid), worker_id, heartbeat)
        self.log.debug('Parsed worker heartbeat: \n{mworker_log}', mworker_log=pprint.pformat(mworker_log.marshal()))

        with self.db.begin(write=True) as txn:
            self.schema.mworker_logs[txn,
                                     (mworker_log.timestamp, mworker_log.node_id, mworker_log.worker_id)] = mworker_log

        ended = time_ns()
        runtime = int(round((ended - started) / 1000000.))

        # taking longer than 250ms means: sth is likely wrong ..
        if runtime > 250:
            self.log.warn(
                'Worker heartbeat excessive processing time {runtime} ms! [node_oid="{node_oid}", worker_id="{worker_id}", timestamp="{timestamp}", sent="{sent}", seq={seq}]',
                runtime=runtime,
                timestamp=mworker_log.timestamp,
                sent=mworker_log.sent,
                seq=mworker_log.seq,
                node_oid=node_oid,
                worker_id=worker_id)
        else:
            self.log.debug(
                'Worker heartbeat processed and stored in {runtime} ms [node_oid="{node_oid}", worker_id="{worker_id}", timestamp="{timestamp}", sent="{sent}", seq={seq}]',
                runtime=runtime,
                timestamp=mworker_log.timestamp,
                sent=mworker_log.sent,
                seq=mworker_log.seq,
                node_oid=node_oid,
                worker_id=worker_id)

        self.log.debug('{func}: completed!', func=hltype(self._on_worker_heartbeat))

    async def _on_node_heartbeat(self, node_authid, heartbeat, details: Optional[CallDetails] = None):
        """
        Receive heartbeats from CF nodes managed by this CFC instances. By default,
         CF nodes will send one hearbeat every 10 seconds.

        :param node_authid: The node WAMP auth ID.
        :type node_authid: str

        :param heartbeat: Node heartbeat.
        :type heartbeat: dict

        :param details: Event details.
        :type details: :class:`autobahn.wamp.types.EventDetails`
        """
        assert type(node_authid) == str
        assert type(heartbeat) == dict
        assert details is None or isinstance(details, EventDetails)

        started = time_ns()

        heartbeat_time = heartbeat.get('timestamp', None)
        assert type(heartbeat_time) == int

        heartbeat_seq = heartbeat.get('seq', None)
        assert type(heartbeat_seq) == int

        heartbeat_pubkey = heartbeat.get('pubkey', None)
        assert type(heartbeat_pubkey) == str and len(heartbeat_pubkey) == 64

        heartbeat_workers = heartbeat.get('workers', {})
        assert type(heartbeat_workers) == dict
        for worker_type in heartbeat_workers:
            # FIXME:
            ALLOWED_WORKER_TYPES = [
                'controller', 'router', 'container', 'guest', 'proxy', 'hostmonitor', 'xbrmm', 'xbr_marketmaker',
                'marketplace'
            ]
            assert worker_type in ALLOWED_WORKER_TYPES, 'invalid worker type "{}" (valid types: {})'.format(
                worker_type, ALLOWED_WORKER_TYPES)
            assert type(heartbeat_workers[worker_type]) == int

        self.log.debug(
            'Heartbeat from node "{node_authid}" with workers {heartbeat_workers} [heartbeat_seq={heartbeat_seq}, time={heartbeat_time}, publisher={publisher}, authid={authid}]',
            node_authid=hlid(node_authid),
            heartbeat_workers=heartbeat_workers,
            heartbeat_seq=hlid(heartbeat_seq),
            heartbeat_time=hlid(heartbeat_time),
            publisher=hlid(details.publisher) if details else None,
            authid=hlid(details.publisher_authid) if details else None)

        self.log.debug('Raw node heartbeat:\n{heartbeat}', heartbeat=pprint.pformat(heartbeat))

        with self.gdb.begin() as txn:
            node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]
            node = self.gschema.nodes[txn, node_oid]

        # currently, nodes are indexed by str-type UUID in the run-time map
        node_oid = str(node_oid)

        if node_oid not in self._nodes:
            self.log.warn(
                'Heartbeat received from "{node_oid}", but node not found in online list (node will now be marked as online)',
                node_oid=node_oid)
            node_shutdown = self._nodes_shutdown.get(node_oid, None)
            if not node_shutdown or (time_ns() - node_shutdown) > (60 * 10**9):

                # create run-time representation of node
                _node = Node(node_oid, heartbeat_seq, heartbeat_time, heartbeat_workers)
                _node.last_activity = Node.LAST_ACTIVITY_HEARTBEAT
                _node.last_active = time_ns()
                self._nodes[node_oid] = _node

                # publish "on_node_ready" management event
                await self._publish_on_node_ready(_node)

                self.log.info('{action} [oid={node_oid}, session={session_id}, status={status}] {func}',
                              action=hl('Success: managed node "{}" is now online'.format(node_authid),
                                        color='green',
                                        bold=True),
                              node_oid=hlid(node_oid),
                              session_id=hlid(details.publisher) if details else None,
                              status=hlval(self._nodes[node_oid].status),
                              func=hltype(self._on_node_heartbeat))
        else:
            self._nodes[node_oid].heartbeat_counter = heartbeat_seq
            self._nodes[node_oid].heartbeat_time = heartbeat_time
            self._nodes[node_oid].heartbeat_workers = heartbeat_workers
            self._nodes[node_oid].last_activity = Node.LAST_ACTIVITY_HEARTBEAT
            self._nodes[node_oid].last_active = time_ns()

            if self._nodes[node_oid].status == 'online':
                self.log.debug('{action} [oid={node_oid}, session={session_id}, status={status}] {func}',
                               action=hl('Ok, managed node "{}" is still alive'.format(node_authid),
                                         color='green',
                                         bold=False),
                               node_oid=hlid(node_oid),
                               session_id=hlid(details.publisher) if details else None,
                               status=hlval(self._nodes[node_oid].status),
                               func=hltype(self._on_node_heartbeat))
            else:
                self.log.info('{action} [oid={node_oid}, session={session_id}, status={status}] {func}',
                              action=hl('Ok, managed node "{}" became alive (again) [status={} -> online]'.format(
                                  node_authid, self._nodes[node_oid].status),
                                        color='yellow',
                                        bold=True),
                              node_oid=hlid(node_oid),
                              session_id=hlid(details.publisher) if details else None,
                              status=hlval(self._nodes[node_oid].status),
                              func=hltype(self._on_node_heartbeat))
                self._nodes[node_oid].status = 'online'

        # heartbeat['authid'] = details.publisher_authid
        heartbeat['authid'] = node_authid
        heartbeat['node_id'] = node_oid
        heartbeat['session'] = details.publisher if details else None

        mrealm_id = node.mrealm_oid
        mnode_log = MNodeLog.parse(mrealm_id, uuid.UUID(node_oid), heartbeat)
        self.log.debug('Parsed node heartbeat:\n{heartbeat}', heartbeat=pprint.pformat(mnode_log.marshal()))

        # this is the pubkey under which an aggregate usage record (see below) will be stored
        if node.pubkey == heartbeat_pubkey:
            with self.db.begin(write=True) as txn:
                self.schema.mnode_logs[txn, (mnode_log.timestamp, mnode_log.node_id)] = mnode_log

            self.log.debug('{msg} [timestamp={timestamp}, node_id={node_id}]',
                           msg=hl('New node HEARTBEAT persisted in database -> checking for pubkey="{}"'.format(
                               node.pubkey),
                                  bold=True),
                           timestamp=hlid(mnode_log.timestamp),
                           node_id=hlid(mnode_log.node_id))

            ended = time_ns()
            runtime = int(round((ended - started) / 1000000.))

            # taking longer than 250ms means: sth is likely wrong ..
            if runtime > 250:
                self.log.warn(
                    'Node heartbeat excessive processing time {runtime} ms! [{node_oid}, timestamp={timestamp}, sent="{sent}", seq={seq}]',
                    runtime=runtime,
                    timestamp=mnode_log.timestamp,
                    sent=mnode_log.sent,
                    seq=mnode_log.seq,
                    node_oid=node_oid)
            else:
                self.log.debug(
                    'Node heartbeat processed and stored in {runtime} ms [node_id={node_oid}, timestamp={timestamp}, sent="{sent}", seq={seq}]',
                    runtime=runtime,
                    timestamp=mnode_log.timestamp,
                    sent=mnode_log.sent,
                    seq=mnode_log.seq,
                    node_oid=node_oid)
        else:
            self.log.warn('heartbeat pubkey does not match pubkey for node matching node_id!')

        self.log.debug('{func}: completed!', func=hltype(self._on_node_heartbeat))

    async def _on_node_shutdown(self, shutdown_info, details: Optional[CallDetails] = None):
        node_authid = shutdown_info.get('node_id', None)
        self.log.info('node "{node_authid}" has shut down: {shutdown_info} {details}',
                      node_authid=node_authid,
                      shutdown_info=shutdown_info,
                      details=details)

        with self.gdb.begin() as txn:
            node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]

        if not node_oid:
            self.log.warn('{func}: unrecognised node shutdown for "{node_authid}" - could not find node for authid',
                          func=hltype(self._on_session_shutdown),
                          node_authid=hlid(node_authid))
            return

        node_oid = str(node_oid)
        if node_oid not in self._nodes:
            self.log.warn('{func}: unrecognised node shutdown for "{node_authid}" - node not in run-time map',
                          func=hltype(self._on_node_shutdown),
                          node_authid=hlid(node_authid))
            return

        if node_oid in self._nodes:
            self._nodes_shutdown[node_oid] = time_ns()

            # mark node as offline in run-time map
            self._nodes[node_oid].status = 'offline'

            # publish management event
            await self._publish_on_node_shutdown(self._nodes[node_oid])

        self.log.debug('{func}: completed!', func=hltype(self._on_node_shutdown))

    async def _on_session_startup(self, session, details: Optional[CallDetails] = None):
        if session.get('authrole') == 'node':
            session_id = session.get('session')
            node_authid = session.get('authid')

            self._sessions[session_id] = node_authid
            with self.gdb.begin() as txn:
                node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]

            # currently, nodes are indexed by str-type UUID in the run-time map
            node_oid = str(node_oid)

            # create run-time representation of node
            if node_oid not in self._nodes:
                _node = Node(node_oid)
                _node.last_activity = Node.LAST_ACTIVITY_STARTED
                _node.last_active = time_ns()
                _node.status = 'online'
                self._nodes[node_oid] = _node

                # publish "on_node_ready" management event
                await self._publish_on_node_ready(_node)

                self.log.info('{action} [oid={node_oid}, session={session_id}, status={status}] {func}',
                              action=hl('Success: managed node "{}" is now online'.format(node_authid),
                                        color='green',
                                        bold=True),
                              node_oid=hlid(node_oid),
                              session_id=hlid(session_id),
                              status=hlval(self._nodes[node_oid].status),
                              func=hltype(self._on_session_startup))

        self.log.debug('{func}: completed!', func=hltype(self._on_session_startup))

    async def _on_session_shutdown(self, session_id, details: Optional[CallDetails] = None):
        node_authid = self._sessions.get(session_id)

        # we are only interested in session closes from management uplinks
        if not node_authid:
            self.log.debug('{func}: unrecognised session close for "{session_id}" - could not map session to authid',
                           func=hltype(self._on_session_shutdown),
                           session_id=hlid(session_id))
            return

        with self.gdb.begin() as txn:
            node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]

        if not node_oid:
            self.log.warn('{func}: unrecognised session close for "{session_id}" - could not find node for authid',
                          func=hltype(self._on_session_shutdown),
                          session_id=hlid(session_id))
            return

        node_oid = str(node_oid)
        if node_oid not in self._nodes:
            self.log.warn('{func}: unrecognised session close for "{session_id}" - node not in run-time map',
                          func=hltype(self._on_session_shutdown),
                          session_id=hlid(session_id))
            return

        self._nodes_shutdown[node_oid] = time_ns()

        # mark node as offline in run-time map
        self._nodes[node_oid].status = 'offline'

        # publish management event
        await self._publish_on_node_shutdown(self._nodes[node_oid])

        self.log.info('{action} [oid={node_oid}, session={session_id}, status={status}] {func}',
                      action=hl('Warning: managed node "{}" became offline'.format(node_authid),
                                color='red',
                                bold=True),
                      node_oid=hlid(node_oid),
                      session_id=hlid(session_id),
                      status=hlval(self._nodes[node_oid].status),
                      func=hltype(self._on_session_shutdown))

    def _check_node_id(self, node_id, status='online'):
        if node_id not in self._nodes:
            raise Exception('no such node: "{node_id}"', node_id=node_id)

        node = self._nodes[node_id]

        if status is not None and node.status != status:
            raise Exception('node "{}" not in status "{}"'.format(node_id, status))

        return node

    def _check_worker_id(self, node_id, worker_id, status='online'):
        self._check_node_id(node_id, status)

    @wamp.register(None, check_types=True)
    def get_status(self, details: Optional[CallDetails] = None) -> dict:
        """
        Get management realm status.

        :returns: Status information object.
        """
        now = utcnow()
        uptime_secs = (iso8601.parse_date(now) - iso8601.parse_date(self._started)).total_seconds()
        uptime_secs_str = humanize.naturaldelta(uptime_secs)
        res = {
            'type': 'management',
            'realm': self._realm,
            'now': utcnow(),
            'started': self._started,
            'uptime': uptime_secs_str,
            'tick': self._tick
        }
        return res

    @wamp.register(None, check_types=True)
    def get_nodes(self,
                  status: Optional[str] = None,
                  return_names: Optional[bool] = None,
                  details: Optional[CallDetails] = None) -> List[str]:
        """
        Returns list of nodes.

        :param status: Filter nodes for this status (``"online"``, ``"offline"``).
        :param return_names: Return node names (``authid``) instead of  object IDs.

        :returns: List of node IDs or node names.
        """
        self.log.info('{func}(status={status}, details.caller_authid={caller_authid})',
                      status=hlval(status),
                      func=hltype(self.get_nodes),
                      caller_authid=hlval(details.caller_authid if details else None))

        with self.gdb.begin() as txn:
            from_key = (self._mrealm_oid, '')
            to_key = (uuid.UUID(int=(int(self._mrealm_oid) + 1)), '')
            node_oids = list(
                self.gschema.idx_nodes_by_authid.select(txn, from_key=from_key, to_key=to_key, return_keys=False))

        if status:
            if status == 'offline':
                res = [node_oid for node_oid in node_oids if str(node_oid) not in self._nodes]
            elif status == 'online':
                res = [
                    node_oid for node_oid in node_oids
                    if (str(node_oid) in self._nodes and self._nodes[str(node_oid)].status == status)
                ]
            else:
                raise Exception('logic error')
        else:
            res = node_oids

        if return_names:
            res_ = []
            with self.gdb.begin() as txn:
                for node_oid in res:
                    node = self.gschema.nodes[txn, node_oid]
                    if node and node.authid:
                        res_.append(node.authid)
        else:
            res_ = [str(node_oid) for node_oid in res]

        return res_

    @wamp.register(None, check_types=True)
    def get_node(self, node_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return information about node. The procedure will raise an `crossbar.error.no_such_object` error
        when no node with the given authid can be found.

        :param node_oid: The object ID of the node to get information for, eg `"5616c7cc-31b5-4021-8cd9-b7769d3f0dd3"`.

        :returns: Node information object.
        """
        try:
            _node_oid = uuid.UUID(node_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid node_oid: {}'.format(str(e)))

        self.log.info('{func}(node_oid="{node_oid}")', node_oid=hlid(_node_oid), func=hltype(self.get_node))

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, _node_oid]
            if not node:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no node with object ID {} in management realm'.format(node_oid))

        node_obj = node.marshal()

        if node_oid in self._nodes:
            node = self._nodes[node_oid]
            node_obj['heartbeat'] = node.heartbeat_counter
            node_obj['timestamp'] = node.heartbeat_time
            node_obj['status'] = node.status
        else:
            node_obj['heartbeat'] = None
            node_obj['timestamp'] = None
            node_obj['status'] = 'offline'

        return node_obj

    @wamp.register(None, check_types=True)
    def get_node_by_authid(self, node_authid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return node information by node (auth)id. The procedure will raise an `crossbar.error.no_such_object` error
        when no node with the given authid can be found.

        :param node_authid: The WAMP authid the node is authenticated under.

        :returns: Node information object.
        """
        self.log.info('{func}(node_authid="{node_authid}")',
                      node_authid=hlid(node_authid),
                      func=hltype(self.get_node_by_authid))

        with self.gdb.begin() as txn:
            node_oid = self.gschema.idx_nodes_by_authid[txn, (self._mrealm_oid, node_authid)]
            if not node_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no node with authid {} in management realm'.format(node_authid))

        node_obj = self.get_node(str(node_oid), details)
        return node_obj

    @inlineCallbacks
    def _init_trace_api(self):
        @inlineCallbacks
        def on_trace_data(node_id, worker_id, trace_id, period, trace_data, details: Optional[CallDetails] = None):
            self.log.debug(
                'Trace "{trace_id}" on node "{node_id}" / worker "{worker_id}":\n\nperiod = {period}\n\ntrace_data = {trace_data}\n\n',
                node_id=node_id,
                worker_id=worker_id,
                trace_id=trace_id,
                period=pprint.pformat(period),
                trace_data=pprint.pformat(trace_data))
            trace = self._traces.get(trace_id, None)
            if trace:

                # beware, we are searching for a list, not a tuple here
                if [node_id, worker_id] in self._traces[trace_id].traced_workers:

                    publish_options = PublishOptions(eligible_authrole=trace.eligible_reader_roles,
                                                     exclude_authrole=trace.exclude_reader_roles,
                                                     acknowledge=True)

                    yield self.publish('crossbarfabriccenter.mrealm.tracing.on_trace_data',
                                       node_id,
                                       worker_id,
                                       trace_id,
                                       period,
                                       trace_data,
                                       options=publish_options)

        yield self.subscribe(on_trace_data, 'crossbarfabriccenter.remote.tracing.on_trace_data',
                             SubscribeOptions(details=True))

        self.log.debug('central tracing API initialized')

    @inlineCallbacks
    @wamp.register(None, check_types=True)
    def get_trace_data(self,
                       trace_id: str,
                       limit: Optional[int] = None,
                       details: Optional[CallDetails] = None) -> Deferred:
        self.log.info('get_trace_data(trace_id="{trace_id}", limit="{limit}")', trace_id=trace_id, limit=limit)

        trace = self._traces.get(trace_id, None)
        if trace:
            if trace.eligible_reader_roles and details:
                if details.caller_authrole not in trace.eligible_reader_roles:
                    raise ApplicationError(u"crossbar.error.no_such_object", "No trace with ID '{}'".format(trace_id))
            if trace.exclude_reader_roles and details:
                if details.caller_authrole in trace.exclude_reader_roles:
                    raise ApplicationError(u"crossbar.error.no_such_object", "No trace with ID '{}'".format(trace_id))
        else:
            raise ApplicationError(u"crossbar.error.no_such_object", "No trace with ID '{}'".format(trace_id))

        dl = []
        for node_id, worker_id in trace.traced_workers:
            node = self._nodes.get(node_id, None)
            if node and node.status == 'online':
                d = self.call('crossbarfabriccenter.remote.tracing.get_trace_data',
                              node_id,
                              worker_id,
                              trace_id,
                              0,
                              limit=limit)
                dl.append(d)
            else:
                dl.append(defer.fail('node not online'))

        trace_data_results = yield DeferredList(dl)
        result: Dict[str, Any] = {}
        for (node_id, worker_id), (success, data) in six.moves.zip(trace.traced_workers, trace_data_results):
            if node_id not in result:
                result[node_id] = {}
            result[node_id][worker_id] = {'success': success, 'data': data}
        returnValue(result)

    @wamp.register(None, check_types=True)
    def get_trace(self, trace_id: str, details: Optional[CallDetails] = None) -> Optional[dict]:
        """
        Get detail information about a previously created trace. When the trace
        doesn't exist, `None` is returned.

        Note: The trace information is only returned when the caller has
        read-access (at least), otherwise `None` is returned (silently).

        :param trace_id: The ID of the trace to retrieve information for.

        :returns: A trace information object.
        """
        trace = self._traces.get(trace_id, None)
        if trace:
            if trace.eligible_reader_roles and details:
                if details.caller_authrole not in trace.eligible_reader_roles:
                    self.log.debug(
                        'get_trace({trace_id}) -> trace found, but not authorized (role "{caller_authrole}" is not eligible)!',
                        trace_id=trace_id,
                        caller_authrole=details.caller_authrole if details else None)
                    return None
            if trace.exclude_reader_roles and details:
                if details.caller_authrole in trace.exclude_reader_roles:
                    self.log.debug(
                        'get_trace({trace_id}) -> trace found, but not authorized (role "{caller_authrole}" is excluded)!',
                        trace_id=trace_id,
                        caller_authrole=details.caller_authrole if details else None)
                    return None
            return trace.marshal()
        else:
            self.log.debug('get_trace({trace_id}) -> no such trace', trace_id=trace_id)
            return None

    @wamp.register(None, check_types=True)
    def get_traces(self, details: Optional[CallDetails] = None) -> List[str]:
        """
        Get IDs of trace defined.

        Note: Only IDs of traces to which the caller has read-access (at least) are returned.

        :returns: List of trace IDs.
        """
        trace_ids = []
        for trace in self._traces.values():
            if trace.eligible_reader_roles and details:
                if details.caller_authrole not in trace.eligible_reader_roles:
                    self.log.info(
                        'get_traces() -> trace "{trace_id}" found, but not authorized (role "{caller_authrole}" is not eligible)!',
                        trace_id=trace.trace_id,
                        caller_authrole=details.caller_authrole)
                    continue
            if trace.exclude_reader_roles and details:
                if details.caller_authrole in trace.exclude_reader_roles:
                    self.log.info(
                        'get_traces() -> trace "{trace_id}" found, but not authorized (role "{caller_authrole}" is excluded)!',
                        trace_id=trace.trace_id,
                        caller_authrole=details.caller_authrole)
                    continue
            trace_ids.append(trace.trace_id)
        return sorted(trace_ids)

    @wamp.register(None, check_types=True)
    def create_trace(self,
                     trace_id: str,
                     traced_workers: List[Tuple[str, str]],
                     trace_options: Optional[Dict] = None,
                     eligible_reader_roles: Optional[List[str]] = None,
                     exclude_reader_roles: Optional[List[str]] = None,
                     details: Optional[CallDetails] = None) -> dict:
        """
        Create a new trace.

        :param trace_id: The ID of the trace to create (must be unique within the management realm).

        :param traced_workers: A list of pairs `(node_id, worker_id)` with node and (router) worker IDs
            on which the trace is to be run.

        :param trace_options: Tracing options for the trace.

        :param eligible_reader_roles: If given, allow read access to the trace only for callers
            authenticated under a WAMP authrole FROM this list - otherwise allow any role (=public)!

        :param exclude_reader_roles: If given, allow read access to the trace only for callers
            authenticated under a WAMP authrole NOT FROM this list - otherwise allow any role (=public)!

        :returns: Trace started information.
        """
        if trace_id in self._traces:
            raise Exception('trace with ID "{}" already exists (status "{}")'.format(
                trace_id, self._traces[trace_id].status))

        status = 'stopped'
        trace = Trace(trace_id, traced_workers, trace_options, eligible_reader_roles, exclude_reader_roles, status)
        self._traces[trace_id] = trace

        trace_created: Dict[str, Any] = {
            # FIXME
        }

        publish_options = PublishOptions(eligible_authrole=trace.eligible_reader_roles,
                                         exclude_authrole=trace.exclude_reader_roles)

        self.publish('{}tracing.on_trace_created'.format(self._uri_prefix),
                     trace_id,
                     trace_created,
                     options=publish_options)

        return trace_created

    @inlineCallbacks
    @wamp.register(None, check_types=True)
    def start_trace(self, trace_id: str, details: Optional[CallDetails] = None) -> Deferred:
        """
        Start a previously created trace.

        :param trace_id: The ID of the trace to start.
        :type trace_id: str

        :returns: dict: Trace started information.
        """
        trace = self._traces.get(trace_id, None)
        if not trace:
            raise Exception('no trace with ID "{}" exists'.format(trace_id))

        if trace.status != 'stopped':
            raise Exception('cannot start trace with ID "{}" currently in status "{}"'.format(trace_id, trace.status))

        trace.status = 'starting'

        publish_options = PublishOptions(eligible_authrole=trace.eligible_reader_roles,
                                         exclude_authrole=trace.exclude_reader_roles)

        self.publish('{}tracing.on_trace_starting'.format(self._uri_prefix), trace_id, options=publish_options)

        traces_started = []
        traces_failed = []

        for node_id, worker_id in trace.traced_workers:
            node = self._nodes.get(node_id, None)
            if node and node.status == 'online':

                try:
                    # stop any currently online traces (we don't want to get data from orphaned traces)
                    traces = yield self.call('crossbarfabriccenter.remote.tracing.get_traces', node_id, worker_id)
                    if trace_id in traces:
                        _trace = yield self.call('crossbarfabriccenter.remote.tracing.get_trace', node_id, worker_id,
                                                 trace_id)
                        if _trace['status'] == 'running':
                            trace_stopped = yield self.call('crossbarfabriccenter.remote.tracing.stop_trace', node_id,
                                                            worker_id, trace_id)
                            self.log.info(
                                'Trace "{trace_id}" on node "{node_id}" / worker "{worker_id}" stopped:\n{trace_stopped}',
                                trace_stopped=trace_stopped,
                                trace_id=trace_id,
                                node_id=node_id,
                                worker_id=worker_id)

                    # start fresh trace
                    trace_started = yield self.call('crossbarfabriccenter.remote.tracing.start_trace',
                                                    node_id,
                                                    worker_id,
                                                    trace_id,
                                                    trace_options=trace.trace_options)
                    self.log.info(
                        'Trace "{trace_id} on node "{node_id}" / worker "{worker_id}" started with options {trace_options}:\n{trace_started}"',
                        node_id=node_id,
                        worker_id=worker_id,
                        trace_id=trace_id,
                        trace_options=trace.trace_options,
                        trace_started=trace_started)

                    traces_started.append(trace_started)
                except Exception as e:
                    self.log.failure()
                    traces_failed.append({
                        'node_id': node_id,
                        'node_status': node.status,
                        'worker_id': worker_id,
                        'error': str(e)
                    })
            else:
                node_status = node.status if node else 'offline'
                self.log.warn('trace "{trace_id}": skipping to start trace on node "{node_id}" in status "{status}"',
                              node_id=node_id,
                              trace_id=trace_id,
                              status=node_status)
                traces_failed.append({'node_id': node_id, 'node_status': node_status, 'worker_id': worker_id})

        trace.status = 'running'

        trace_started = {
            'started': traces_started,
            'failed': traces_failed,
        }

        self.publish('{}tracing.on_trace_started'.format(self._uri_prefix),
                     trace_id,
                     trace_started,
                     options=publish_options)

        returnValue(trace_started)

    @inlineCallbacks
    @wamp.register(None, check_types=True)
    def stop_trace(self, trace_id: str, details: Optional[CallDetails] = None) -> Deferred:
        """
        Stop a running trace.

        :param trace_id: The ID of the trace to stop.

        :returns: Trace stopped information.
        """
        trace = self._traces.get(trace_id, None)
        if not trace:
            raise Exception('no trace with ID "{}" exists'.format(trace_id))

        if trace.status not in ['running', 'stopping_failed']:
            raise Exception('cannot stop trace with ID "{}" currently in status "{}"'.format(trace_id, trace.status))

        trace.status = 'stopping'

        publish_options = PublishOptions(eligible_authrole=trace.eligible_reader_roles,
                                         exclude_authrole=trace.exclude_reader_roles)

        self.publish('{}tracing.on_trace_stopping'.format(self._uri_prefix), trace_id, options=publish_options)

        traces_stopped = []
        traces_failed = []

        for node_id, worker_id in trace.traced_workers:
            node = self._nodes.get(node_id, None)
            if node and node.status == 'online':

                try:
                    traces = yield self.call('crossbarfabriccenter.remote.tracing.get_traces', node_id, worker_id)
                    if trace_id in traces:
                        _trace = yield self.call('crossbarfabriccenter.remote.tracing.get_trace', node_id, worker_id,
                                                 trace_id)
                        if _trace['status'] == 'running':
                            trace_stopped = yield self.call('crossbarfabriccenter.remote.tracing.stop_trace', node_id,
                                                            worker_id, trace_id)
                            traces_stopped.append(trace_stopped)
                            self.log.info(
                                'Trace "{trace_id}" on node "{node_id}" / worker "{worker_id}" stopped:\n{trace_stopped}',
                                trace_stopped=trace_stopped,
                                trace_id=trace_id,
                                node_id=node_id,
                                worker_id=worker_id)
                except Exception as e:
                    self.log.failure()
                    traces_failed.append({
                        'node_id': node_id,
                        'node_status': node.status,
                        'worker_id': worker_id,
                        'error': str(e)
                    })
            else:
                node_status = node.status if node else 'offline'
                self.log.warn('trace "{trace_id}": skipping to stop trace on node "{node_id}" in status "{status}"',
                              node_id=node_id,
                              trace_id=trace_id,
                              status=node_status)
                traces_failed.append({'node_id': node_id, 'node_status': node_status, 'worker_id': worker_id})

        if len(traces_failed):
            trace.status = 'stopping_failed'
        else:
            trace.status = 'stopped'

        trace_stopped = {
            'stopped': traces_stopped,
            'failed': traces_failed,
        }

        self.publish('{}tracing.on_trace_stopped'.format(self._uri_prefix),
                     trace_id,
                     trace_stopped,
                     options=publish_options)

        returnValue(trace_stopped)

    @wamp.register(None, check_types=True)
    def delete_trace(self, trace_id: str, details: Optional[CallDetails] = None) -> dict:
        """
        Delete a previously created (and currently stopped) trace.

        :param trace_id: The ID of the trace to delete.
        :returns: Trace deletion information.
        """
        trace = self._traces.get(trace_id, None)
        if not trace:
            raise Exception('no trace with ID "{}" exists'.format(trace_id))

        if trace.status not in ['stopped', 'stopping_failed']:
            raise Exception('cannot delete trace with ID "{}" currently in status "{}"'.format(trace_id, trace.status))

        trace_deleted: Dict[str, Any] = {
            # FIXME
        }

        publish_options = PublishOptions(eligible_authrole=trace.eligible_reader_roles,
                                         exclude_authrole=trace.exclude_reader_roles)

        self.publish('{}tracing.on_trace_deleted'.format(self._uri_prefix),
                     trace_id,
                     trace_deleted,
                     options=publish_options)

        del self._traces[trace_id]

        return trace_deleted



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from typing import Union
from txaio import make_logger, as_future

from autobahn import util
from autobahn.wamp.types import Accept, Deny, HelloDetails, Challenge

from crossbar.router.auth.pending import PendingAuth
from crossbar._util import hlid, hltype, hlval

__all__ = (
    'PendingAuthAnonymous',
    'PendingAuthAnonymousProxy',
)


class PendingAuthAnonymous(PendingAuth):
    """
    Pending authentication information for WAMP-Anonymous authentication.
    """

    log = make_logger()

    AUTHMETHOD = 'anonymous'

    def hello(self, realm: str, details: HelloDetails) -> Union[Accept, Deny, Challenge]:
        self.log.info(
            '{func}(realm={realm}, details.realm={authrealm}, details.authid={authid}, details.authrole={authrole}) [config={config}]',
            func=hltype(self.hello),
            realm=hlid(realm),
            authrealm=hlid(details.realm),
            authid=hlid(details.authid),
            authrole=hlid(details.authrole),
            config=self._config)

        # remember the realm the client requested to join (if any)
        self._realm = realm

        self._authid = self._config.get('authid', util.generate_serial_number())

        self._session_details['authmethod'] = 'anonymous'
        self._session_details['authextra'] = details.authextra

        # WAMP-anonymous "static"
        if self._config['type'] == 'static':

            self._authprovider = 'static'

            # FIXME: if cookie tracking is enabled, set authid to cookie value
            # self._authid = self._transport._cbtid

            principal = {
                'realm': realm,
                'authid': self._authid,
                'role': self._config.get('role', 'anonymous'),
                'extra': details.authextra
            }

            error = self._assign_principal(principal)
            if error:
                return error

            return self._accept()

        # WAMP-Ticket "dynamic"
        elif self._config['type'] == 'dynamic':

            self._authprovider = 'dynamic'

            init_d = as_future(self._init_dynamic_authenticator)

            def init(result):
                if result:
                    return result

                d = self._authenticator_session.call(self._authenticator, self._realm, self._authid,
                                                     self._session_details)

                def on_authenticate_ok(_principal):
                    _error = self._assign_principal(_principal)
                    if _error:
                        return _error

                    return self._accept()

                def on_authenticate_error(err):
                    return self._marshal_dynamic_authenticator_error(err)

                d.addCallbacks(on_authenticate_ok, on_authenticate_error)

                return d

            init_d.addBoth(init)
            return init_d

        else:
            # should not arrive here, as config errors should be caught earlier
            return Deny(message='invalid authentication configuration (authentication type "{}" is unknown)'.format(
                self._config['type']))


class PendingAuthAnonymousProxy(PendingAuthAnonymous):
    """
    Pending Anonymous authentication with additions for proxy
    """

    log = make_logger()
    AUTHMETHOD = 'anonymous-proxy'

    def hello(self, realm, details):
        self.log.debug('{func}(realm="{realm}", details={details})',
                       func=hltype(self.hello),
                       realm=hlval(realm),
                       details=details)
        extra = details.authextra or {}

        for attr in ['proxy_authid', 'proxy_authrole', 'proxy_realm']:
            if attr not in extra:
                return Deny(message='missing required attribute {}'.format(attr))

        realm = extra['proxy_realm']
        details.authid = extra['proxy_authid']
        details.authrole = extra['proxy_authrole']
        details.authextra = extra.get('proxy_authextra', None)

        # remember the realm the client requested to join (if any)
        self._realm = realm
        self._authid = details.authid
        self._authrole = details.authrole
        self._session_details['authmethod'] = self.AUTHMETHOD
        self._session_details['authextra'] = details.authextra
        self._authprovider = 'static'

        # FIXME: if cookie tracking is enabled, set authid to cookie value
        # self._authid = self._transport._cbtid

        principal = {'realm': realm, 'authid': details.authid, 'role': details.authrole, 'extra': details.authextra}

        error = self._assign_principal(principal)
        if error:
            return error

        return self._accept()



from autobahn.twisted.util import sleep


def setup_auth(session, details):
    def register(_):
        def authenticate(*args, **kw):
            print("test.authenticate: {} {}".format(args, kw))
            return {
                "allow": False,
                "role": "role0",
                # corresponds to a private-key of all "a"s
                "pubkey": "e734ea6c2b6257de72355e472aa05a4c487e6b463c029ed306df2f01b5636b58",
            }

        reg = session.register(authenticate, u"test.authenticate")
        print("registered: {}\n\n\n".format(reg))
        return reg

    # we wait here to simulate slow startup of the authenticator
    d = sleep(5)
    d.addCallback(register)
    return d



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import re

from txaio import make_logger

from autobahn.wamp.types import RegisterOptions, SubscribeOptions

__all__ = ('RemoteWampApi', )


class RemoteWampApi(object):
    """
    Exposes the WAMP meta API on a router realm from a remote node
    on CFC. This is only available if the router realm is started
    with option.bridge_meta_api == True.

    Events are published locally on the node WAMP meta API like

    [1] wamp.session.on_join
            (session_id, session_details)

    and (if bridged), on the local node management API

    [2] crossbar.worker.<worker_id>.realm.<realm_id>.root.wamp.session.on_join
            (session_id, session_details)

    which then get forward to CFC as

    [3] crossbarfabriccenter.node.<node_id>.worker.<worker_id>.realm.<realm_id>.root.
            (session_id, session_details)

    which is then republished by CFC as

    [4] crossbarfabriccenter.remote.realm.meta.wamp.session.on_join
            (node_id, worker_id, realm_id, session_id, session_details)

    For example:

    [1] wamp.session.on_join
    [2] crossbar.worker.worker-001.realm.realm-001.root.wamp.session.on_join
    [3] crossbarfabriccenter.node.cf1.worker.worker-001.realm.realm-001.root.wamp.session.on_join
    [4] crossbarfabriccenter.remote.realm.meta.wamp.session.on_join
    """

    log = make_logger()

    PREFIX = 'crossbarfabriccenter.remote.realm.meta.'

    def register(self, session):

        try:

            def forward(node_id, worker_id, realm_id, *args, **kwargs):
                details = kwargs.pop('details', None)
                proc = details.procedure[len(self.PREFIX):]
                uri = 'crossbarfabriccenter.node.{node_id}.worker.{worker_id}.realm.{realm_id}.root.{proc}'.format(
                    node_id=node_id, worker_id=worker_id, realm_id=realm_id, proc=proc)
                self.log.debug(
                    'Forwarding CFC remote WAMP meta API call <{proc}> on management realm "{mrealm}" to node_id "{node_id}", worker_id "{worker_id}", realm_id "{realm_id}"',
                    node_id=node_id,
                    worker_id=worker_id,
                    realm_id=realm_id,
                    proc=details.procedure,
                    mrealm=session._realm)
                return session.call(uri, *args, **kwargs)

            reg = session.register(forward,
                                   self.PREFIX,
                                   options=RegisterOptions(match='prefix', details_arg='details'))
            return [reg]

        except:
            self.log.failure()
            return []

    def subscribe(self, session):

        remote_uri = 'crossbarfabriccenter.node..worker..realm..root.'
        remote_uri_regex = r'^crossbarfabriccenter.node.([a-z0-9][a-z0-9_\-]*).worker.([a-z0-9][a-z0-9_\-]*).realm.([a-z0-9][a-z0-9_\-]*).root.(\S*)$'
        remote_uri_pat = re.compile(remote_uri_regex)

        def forward(*args, **kwargs):
            try:
                details = kwargs.pop('details', None)

                if details:
                    match = remote_uri_pat.match(details.topic)
                    if match and len(match.groups()) == 4:
                        node_id, worker_id, realm_id, suffix_uri = match.groups()

                        # reverse our hack: see crossbar.router.service.RouterServiceSession.publish
                        suffix_uri = suffix_uri.replace('-', '.')

                        local_uri = '{}{}'.format(self.PREFIX, suffix_uri)
                        self.log.debug('RemoteWampApi.forward("{topic}") -> "{local_uri}"',
                                       topic=details.topic,
                                       local_uri=local_uri)
                        return session.publish(local_uri, node_id, worker_id, realm_id, *args, **kwargs)

                # should not arrive here
                session.log.warn(
                    'received unexpected WAMP meta event to forward for management API: details={details}',
                    details=details)
            except:
                session.log.failure()

        sub = session.subscribe(forward, remote_uri, SubscribeOptions(match='wildcard', details=True))

        return [sub]



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import copy
import json
import pkg_resources

import pytest

from crossbar._util import merge_config, _deep_merge_object
from crossbar.master import Personality as MasterPersonality


@pytest.fixture(scope='function')
def master_personality():
    return MasterPersonality


@pytest.fixture(scope='function')
def master_config():
    # built-in config
    # filename = pkg_resources.resource_filename('crossbar', 'master/node/config.json')

    # "copied" built-in config for testing
    filename = pkg_resources.resource_filename('crossbar', 'test/test_config.json')

    with open(filename) as f:
        config = json.load(f)
        MasterPersonality.check_config(MasterPersonality, config)
    return config


def test_deep_merge_map1():
    """
    Test merging two empty maps.
    """
    a = {}
    b = {}
    c = _deep_merge_object(a, b)
    assert c == {}


def test_deep_merge_map2():
    """
    Test merging an empty map into a non-empty map.
    """
    a = {'foo': 23}
    b = {}
    c = _deep_merge_object(a, b)
    assert c == a


def test_deep_merge_map3():
    """
    Test merging a non-empty map into an empty map.
    """
    a = {}
    b = {'foo': 23}
    c = _deep_merge_object(a, b)
    assert c == b


def test_deep_merge_map4():
    """
    Test removing a key from a non-empty map.
    """
    a = {'foo': 23}
    b = {'foo': None}
    c = _deep_merge_object(a, b)
    assert c == {}


def test_deep_merge_map5():
    """
    Test removing a non-existent key from a map.
    """
    a = {'foo': 23}
    b = {'bar': None}
    c = _deep_merge_object(a, b)
    assert c == a


def test_deep_merge_map6():
    """
    Test adding a new key to a map.
    """
    a = {'foo': 23}
    b = {'bar': 666}
    c = _deep_merge_object(a, b)
    assert c == {'foo': 23, 'bar': 666}


def test_deep_merge_map7():
    """
    Test "expanding" a nested map with an empty map.
    """
    a = {'foo': 23, 'bar': {'baz': 666}}
    b = {'bar': {}}
    c = _deep_merge_object(a, b)
    assert c == {'foo': 23, 'bar': {'baz': 666}}


def test_deep_merge_map8():
    """
    Test removing an existing key in a nested map.
    """
    a = {'foo': 23, 'bar': {'baz': 666}}
    b = {'bar': {'baz': None}}
    c = _deep_merge_object(a, b)
    assert c == {'foo': 23, 'bar': {}}


def test_deep_merge_map9():
    """
    Test replacing an existing key in a nested map.
    """
    a = {'foo': 23, 'bar': {'baz': 666}}
    b = {'bar': {'baz': 888}}
    c = _deep_merge_object(a, b)
    assert c == {'foo': 23, 'bar': {'baz': 888}}


def test_deep_merge_map10():
    """
    Test expanding a nested map with a new key.
    """
    a = {'foo': 23, 'bar': {'baz': 666}}
    b = {'bar': {'moo': 888}}
    c = _deep_merge_object(a, b)
    assert c == {'foo': 23, 'bar': {'baz': 666, 'moo': 888}}


def test_deep_merge_list1():
    """
    Test merging two empty lists.
    """
    a = []
    b = []
    c = _deep_merge_object(a, b)
    assert c == []


def test_deep_merge_list2():
    """
    Test copying a list.
    """
    a = [1, 2, 3]
    b = ['COPY', 'COPY', 'COPY']
    c = _deep_merge_object(a, b)
    assert c == [1, 2, 3]


def test_deep_merge_list3():
    """
    Test copying a list.
    """
    a = [1, 2, 3]
    b = [None, None, None]
    c = _deep_merge_object(a, b)
    assert c == []


def test_deep_merge_list4():
    """
    Test expanding a list at the tail.
    """
    a = [1, 2, 3]
    b = ['COPY', 'COPY', 'COPY', 4]
    c = _deep_merge_object(a, b)
    assert c == [1, 2, 3, 4]


def test_deep_merge_list5():
    """
    Test replacing a list.
    """
    a = [1, 2, 3]
    b = [None, None, None, 4, 5]
    c = _deep_merge_object(a, b)
    assert c == [4, 5]


def test_deep_merge_list6():
    """
    Test partially dropping items, copying and appending items.
    """
    a = [1, 2, 3]
    b = [None, 'COPY', None, 4]
    c = _deep_merge_object(a, b)
    assert c == [2, 4]


def test_deep_merge_list7():
    """
    Test partially dropping items, copying and appending items.
    """
    a = [1, 2, 3]
    b = [0, None, 'COPY', 4, 5]
    c = _deep_merge_object(a, b)
    assert c == [0, 3, 4, 5]


def test_deep_merge_list8():
    """
    """
    a = [{'foo': 23}, {'bar': 666}, {'baz': 42}]
    b = ['COPY', 'COPY', 'COPY']
    c = _deep_merge_object(a, b)
    assert c == a


def test_deep_merge_list9():
    """
    """
    a = [{'foo': 23}, {'bar': 666}, {'baz': 42}]
    b = ['COPY', None, 'COPY']
    c = _deep_merge_object(a, b)
    assert c == [{'foo': 23}, {'baz': 42}]


def test_deep_merge_list10():
    """
    """
    a = [{'foo': 23}, {'bar': 666}, {'baz': 42}]
    b = [None, 'COPY', 'COPY', {'moo': 99}]
    c = _deep_merge_object(a, b)
    assert c == [{'bar': 666}, {'baz': 42}, {'moo': 99}]


def test_deep_merge_list11():
    """
    """
    a = [{'foo': 23}, {'bar': 666}]
    b = [{'foo': 33}, {'goo': 17}]
    c = _deep_merge_object(a, b)
    assert c == [{'foo': 33}, {'bar': 666, 'goo': 17}]


def test_deep_merge_complex1():
    """
    """
    a = [{'foo': [1, 2, 3]}, {'bar': [{'baz': 666}, [4, 5, 6]]}]
    b = [{'foo': [None, 5, 'COPY', 9]}, {'bar': [{'goo': 777}, 'COPY']}]
    c = _deep_merge_object(a, b)
    assert c == [{'foo': [5, 3, 9]}, {'bar': [{'baz': 666, 'goo': 777}, [4, 5, 6]]}]


def test_merge_empty(master_personality, master_config):
    cfg = merge_config(master_config, {})
    assert cfg == master_config


def test_merge_same(master_personality, master_config):
    same_config = copy.deepcopy(master_config)
    cfg = merge_config(master_config, same_config)

    assert cfg == master_config


def make_test_func(cfg_name):
    def test_merge_config(master_personality, master_config):
        filename = pkg_resources.resource_filename('crossbar', 'tests/test_{}.json'.format(cfg_name))
        with open(filename) as f:
            test_config = json.load(f)

            # a mergeable override-config does NOT need to be valid in itself
            # MasterPersonality.check_config(MasterPersonality, test_config)

        filename = pkg_resources.resource_filename('crossbar', 'tests/test_{}_merged.json'.format(cfg_name))
        with open(filename) as f:
            test_config_merged = json.load(f)
            MasterPersonality.check_config(MasterPersonality, test_config_merged)

        cfg = merge_config(master_config, test_config)

        assert cfg == test_config_merged

    return test_merge_config


current_module = __import__(__name__)

for cfg_name in ['config1', 'config2', 'config3']:
    setattr(current_module, 'test_merge_{}'.format(cfg_name), make_test_func(cfg_name))

if __name__ == '__main__':
    import sys
    sys.setrecursionlimit(10000)

    # this is the base config into the test configs will be merged
    filename = pkg_resources.resource_filename('crossbar', 'tests/test_config.json')
    with open(filename) as f:
        _master_config = json.load(f)

    same_config = copy.deepcopy(_master_config)
    same_config = {
        'controller': {
            'id': 'mynode1',
        },
        'workers': [
            'COPY',
            'COPY',
        ]
    }
    same_config = {
        'controller': {
            'id': 'mynode1',
        }
    }
    cfg = merge_config(_master_config, same_config)
    assert cfg == _master_config



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from typing import Dict, Any

import txaio
from txaio import make_logger

from autobahn.wamp.types import Deny, TransportDetails

from crossbar.router.auth.pending import PendingAuth
from crossbar.interfaces import IRealmContainer, IPendingAuth

__all__ = ('PendingAuthTLS', )


class PendingAuthTLS(PendingAuth):
    """
    Pending WAMP-TLS authentication.
    """

    AUTHMETHOD = 'tls'

    log = make_logger()

    def __init__(self, pending_session_id: int, transport_details: TransportDetails, realm_container: IRealmContainer,
                 config: Dict[str, Any]):
        super(PendingAuthTLS, self).__init__(
            pending_session_id,
            transport_details,
            realm_container,
            config,
        )

        # https://tools.ietf.org/html/rfc5056
        # https://tools.ietf.org/html/rfc5929
        # https://www.ietf.org/proceedings/90/slides/slides-90-uta-0.pdf
        self._channel_id = transport_details.channel_id.get('tls-unique',
                                                            None) if transport_details.channel_id else None
        self._peer_cert = transport_details.peer_cert

        # for static-mode, the config has principals as a dict indexed
        # by authid, but we need the reverse map: cert-sha1 -> principal
        self._cert_sha1_to_principal = None
        if self._config['type'] == 'static':
            self._cert_sha1_to_principal = {}
            if 'principals' in self._config:
                for authid, principal in self._config['principals'].items():
                    self._cert_sha1_to_principal[principal['certificate-sha1']] = {
                        'authid': authid,
                        'role': principal['role']
                    }

    def hello(self, realm, details):

        # we must have a client TLS certificate to continue
        if not self._peer_cert:
            return Deny(message='client did not send a TLS client certificate')

        # remember the realm the client requested to join (if any)
        self._realm = realm

        # remember the authid the client wants to identify as (if any)
        self._authid = details.authid

        # use static principal database from configuration
        if self._config['type'] == 'static':

            self._authprovider = 'static'

            client_cert_sha1 = self._peer_cert['sha1']
            if client_cert_sha1 in self._cert_sha1_to_principal:
                principal = self._cert_sha1_to_principal[client_cert_sha1]
                error = self._assign_principal(principal)
                if error:
                    return error
                return self._accept()
            else:
                return Deny(message='no principal with authid "{}" exists'.format(client_cert_sha1))

        # use configured procedure to dynamically get a ticket for the principal
        elif self._config['type'] == 'dynamic':

            self._authprovider = 'dynamic'

            init_d = txaio.as_future(self._init_dynamic_authenticator)

            def init(result):
                if result:
                    return result

                self._session_details['authmethod'] = self._authmethod  # from AUTHMETHOD, via base
                self._session_details['authextra'] = details.authextra

                d = self._authenticator_session.call(self._authenticator, realm, details.authid, self._session_details)

                def on_authenticate_ok(_principal):
                    _error = self._assign_principal(_principal)
                    if _error:
                        return _error

                    # FIXME: not sure about this .. TLS is a transport-level auth mechanism .. so forward
                    # self._transport._authid = self._authid
                    # self._transport._authrole = self._authrole
                    # self._transport._authmethod = self._authmethod
                    # self._transport._authprovider = self._authprovider
                    # self._transport._authextra = self._authextra

                    return self._accept()

                def on_authenticate_error(err):
                    return self._marshal_dynamic_authenticator_error(err)

                d.addCallbacks(on_authenticate_ok, on_authenticate_error)
                return d

            init_d.addBoth(init)
            return init_d

        else:
            # should not arrive here, as config errors should be caught earlier
            return Deny(message='invalid authentication configuration (authentication type "{}" is unknown)'.format(
                self._config['type']))

    def authenticate(self, signature):
        # should not arrive here!
        raise Exception("internal error (WAMP-TLS does not implement AUTHENTICATE)")


IPendingAuth.register(PendingAuthTLS)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import uuid
from collections import deque
from typing import Dict, List, Any, Optional, Tuple

import zlmdb
import numpy as np

from txaio import use_twisted  # noqa
from txaio import make_logger, time_ns

from twisted.internet.defer import inlineCallbacks

from autobahn.util import hltype, hlval
from autobahn.twisted.util import sleep
from autobahn.wamp import message
from autobahn.wamp.types import CloseDetails, SessionDetails, TransportDetails
from autobahn.wamp.message import Publish
from autobahn.wamp.interfaces import ISession

from crossbar.router.observation import UriObservationMap
from crossbar.router.realmstore import QueuedCall
from crossbar.interfaces import IRealmStore

import cfxdb
from cfxdb.realmstore import RealmStore, Publication

__all__ = ('RealmStoreDatabase', )


class RealmStoreDatabase(object):
    """
    Database-backed realm store.
    """
    log = make_logger()

    GLOBAL_HISTORY_LIMIT = 100
    """
    The global history limit, in case not overridden.
    """

    STORE_TYPE = 'cfxdb'

    def __init__(self, personality, factory, config):
        """

        :param personality:
        :param factory:
        :param config: Realm store configuration item.
        """
        from twisted.internet import reactor

        self._reactor = reactor
        self._personality = personality
        self._factory = factory

        dbpath = config.get('path', None)
        assert type(dbpath) == str

        maxsize = config.get('maxsize', 128 * 2**20)
        assert type(maxsize) == int
        # allow maxsize 128kiB to 128GiB
        assert maxsize >= 128 * 1024 and maxsize <= 128 * 2**30

        readonly = config.get('readonly', False)
        assert type(readonly) == bool

        sync = config.get('sync', True)
        assert type(sync) == bool

        self._config = config

        self._type = self._config.get('type', None)
        assert self._type == self.STORE_TYPE

        self._db = zlmdb.Database(dbpath=dbpath, maxsize=maxsize, readonly=readonly, sync=sync)
        self._db.__enter__()
        self._schema = RealmStore.attach(self._db)

        self._running = False
        self._process_buffers_thread = None

        self._max_buffer = config.get('max-buffer', 10000)
        self._buffer_flush = config.get('buffer-flush', 200)
        self._buffer = []
        self._log_counter = 0

        # map: registration.id -> deque( (session, call, registration, authorization) )
        self._queued_calls = {}

        self.log.info(
            '{func} realm store initialized (type="{stype}", dbpath="{dbpath}", maxsize={maxsize}, '
            'readonly={readonly}, sync={sync})',
            func=hltype(self.__init__),
            stype=hlval(self._type),
            dbpath=dbpath,
            maxsize=maxsize,
            readonly=readonly,
            sync=sync)

    def type(self) -> str:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.type`
        """
        return self._type

    def is_running(self) -> bool:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.is_running`
        """
        return self._running

    @inlineCallbacks
    def start(self):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.start`
        """
        if self._running:
            raise RuntimeError('store is already running')
        else:
            self.log.info(
                '{func} starting realm store type="{stype}"',
                func=hltype(self.start),
                stype=hlval(self._type),
            )

        self._buffer = []
        self._log_counter = 0
        self._running = True
        self._process_buffers_thread = yield self._reactor.callInThread(self._process_buffers)
        self.log.info('{func} realm store ready!', func=hltype(self.start))

    def stop(self):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.stop`
        """
        if not self._running:
            raise RuntimeError('store is not running')
        else:
            self.log.info('{func} stopping realm store', func=hltype(self.start))

        self._running = False

        # FIXME: wait for _process_buffers_thread

        self._buffer = []
        self._log_counter = 0

    @inlineCallbacks
    def _process_buffers(self):
        self.log.debug('ZLMDB buffer background writer starting')
        while self._running:
            cnt, errs, duration_ms = self._process_buffer()
            if cnt > 0:
                self.log.debug('ZLMDB buffer background writer processed {cnt} records in {duration_ms} ms',
                               cnt=cnt,
                               duration_ms=duration_ms)
            if errs > 0:
                self.log.warn('ZLMDB buffer background writer encountered {errs} errors', errs=errs)
            if duration_ms < self._buffer_flush:
                sleep_ms = int(self._buffer_flush - duration_ms)
            else:
                sleep_ms = self._buffer_flush
            self.log.debug('Throttling buffer background writer (sleeping {sleep_ms} ms)', sleep_ms=sleep_ms)
            yield sleep(float(sleep_ms) / 1000.)
        self.log.debug('ZLMDB buffer background writer ended')

    def _process_buffer(self):
        buffer = self._buffer
        self._buffer = []
        self.log.debug('Processing {cnt} buffered records', cnt=len(buffer))
        cnt = 0
        errs = 0
        started = time_ns()
        with self._db.begin(write=True) as txn:
            for rec in buffer:
                func = rec[0]
                args = rec[1:]
                try:
                    func(txn, *args)
                    cnt += 1
                except:
                    self.log.failure()
                    errs += 1
            ended = time_ns()
        duration_ms = int((ended - started) / 1000000)
        return cnt, errs, duration_ms

    def store_session_joined(self, session: ISession, details: SessionDetails):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_session_joined`
        """
        self.log.debug('{func} new session joined session={session}, details={details}',
                       func=hltype(self.store_session_joined),
                       session=session,
                       details=details)

        ses = cfxdb.realmstore.Session()
        ses.oid = uuid.uuid4()
        ses.joined_at = np.datetime64(time_ns(), 'ns')
        ses.session = details.session
        ses.realm = details.realm
        ses.authid = details.authid
        ses.authrole = details.authrole
        ses.authmethod = details.authmethod
        ses.authprovider = details.authprovider
        ses.authextra = details.authextra

        # the client frontend transport, both in router-based and proxy-router setups
        ses.transport = details.transport.marshal()

        # in proxy-router setups, this transport is the proxy-to-router transport,
        # whereas in plain router-based setups, this is the same as above.
        ptd = session._transport.transport_details.marshal()

        # FIXME: we should have a better way to recognize proxy-router setups
        if ptd != ses.transport:
            ses.proxy_transport = ptd

        # FIXME: fill with the proxy-to-router authentication of the proxy itself
        ses.proxy_node_oid = None
        ses.proxy_node_authid = None
        ses.proxy_worker_name = None
        ses.proxy_worker_pid = None

        self._buffer.append([self._store_session_joined, ses])

    def _store_session_joined(self, txn: zlmdb.Transaction, ses: cfxdb.realmstore.Session):

        # FIXME: use idx_sessions_by_session_id to check there is no session with (session_id, joined_at) yet

        self._schema.sessions[txn, ses.oid] = ses

        cnt = self._schema.sessions.count(txn)
        self.log.info('{func} database record inserted [total={total}] session={session}',
                      func=hltype(self._store_session_joined),
                      total=hlval(cnt),
                      session=ses)

    def store_session_left(self, session: ISession, details: CloseDetails):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_session_left`
        """
        self.log.debug('{func} session left session={session}, details={details}',
                       func=hltype(self.store_session_left),
                       session=session,
                       details=details)

        self._buffer.append([self._store_session_left, session, details])

    def _store_session_left(self, txn: zlmdb.Transaction, session: ISession, details: CloseDetails):

        # FIXME: apparently, session ID is already erased at this point:(
        _session_id = session._session_id

        # FIXME: move left_at to autobahn.wamp.types.CloseDetails
        _left_at = np.datetime64(time_ns(), 'ns')

        # lookup session by WAMP session ID and find the most recent session
        # according to joined_at timestamp
        session_obj = None
        _from_key = (_session_id, np.datetime64(0, 'ns'))
        _to_key = (_session_id, np.datetime64(time_ns(), 'ns'))
        for session_oid in self._schema.idx_sessions_by_session_id.select(txn,
                                                                          from_key=_from_key,
                                                                          to_key=_to_key,
                                                                          reverse=True,
                                                                          return_keys=False,
                                                                          return_values=True):
            session_obj = self._schema.sessions[txn, session_oid]

            # if we have an index, that index must always resolve to an indexed record
            assert session_obj

            # we only want the most recent session
            break

        if session_obj:
            # FIXME: also store other CloseDetails attributes
            session_obj.left_at = _left_at

            self.log.info('{func} database record session={session} updated: left_at={left_at}',
                          func=hltype(self._store_session_left),
                          left_at=hlval(_left_at),
                          session=hlval(_session_id))
        else:
            self.log.warn('{func} could not update database record for session={session}: record not found!',
                          func=hltype(self._store_session_left),
                          session=hlval(_session_id))

    def get_session_by_session_id(self,
                                  session_id: int,
                                  joined_before: Optional[int] = None) -> Optional[Dict[str, Any]]:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_session_by_session_id`
        """
        if joined_before:
            _joined_before = np.datetime64(joined_before, 'ns')
        else:
            _joined_before = np.datetime64(time_ns(), 'ns')
        _from_key = (session_id, np.datetime64(0, 'ns'))
        _to_key = (session_id, _joined_before)

        # check if we have a record store for the session
        session: Optional[cfxdb.realmstore.Session] = None
        with self._db.begin() as txn:
            # lookup session by WAMP session ID and find the most recent session
            # according to joined_at timestamp
            for session_oid in self._schema.idx_sessions_by_session_id.select(txn,
                                                                              from_key=_from_key,
                                                                              to_key=_to_key,
                                                                              reverse=True,
                                                                              return_keys=False,
                                                                              return_values=True):
                session = self._schema.sessions[txn, session_oid]

                # if we have an index, that index must always resolve to an indexed record
                assert session

                # we only want the most recent session
                break

        if session:
            # extract info from database table to construct session details and return
            td = TransportDetails.parse(session.transport)
            sd = SessionDetails(
                realm=session.realm,
                session=session.session,
                authid=session.authid,
                authrole=session.authrole,
                authmethod=session.authmethod,
                authprovider=session.authprovider,
                authextra=session.authextra,
                # FIXME
                serializer=None,
                resumed=False,
                resumable=False,
                resume_token=None,
                transport=td)
            res = sd.marshal()
            return res
        else:
            return None

    def get_sessions_by_authid(self, authid: str) -> Optional[List[Tuple[str, int]]]:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_sessions_by_authid`
        """

    def attach_subscription_map(self, subscription_map: UriObservationMap):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.attach_subscription_map`
        """
        for sub in self._config.get('event-history', []):
            uri = sub['uri']
            match = sub.get('match', u'exact')
            # observation, was_already_observed, was_first_observer
            subscription_map.add_observer(self, uri=uri, match=match)
            # subscription_id = observation.id

    def store_event(self, session: ISession, publication_id: int, publish: Publish):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_event`
        """
        # FIXME: builtins.AssertionError: invalid type <class 'crossbar.router.service.RouterServiceAgent'> for "session"
        # assert isinstance(session, RouterSession), 'invalid type {} for "session"'.format(type(session))
        assert type(publication_id) == int, 'invalid type {} for "publication_id"'.format(type(publication_id))
        assert isinstance(publish, message.Publish), 'invalid type {} for "publish"'.format(type(publish))

        self._buffer.append([self._store_event, session, publication_id, publish])

    def _store_event(self, txn, session, publication_id, publish):

        pub = self._schema.publications[txn, publication_id]
        if pub:
            raise Exception('duplicate event for publication_id={}'.format(publication_id))

        # FIXME: use idx_sessions_by_session_id
        # ses = self._schema.sessions[txn, session._session_id] if session._session_id else None
        # if not ses:
        #     self.log.info(
        #         'session {session} (realm={realm}, authid={authid}, authrole={authrole}) not in store - '
        #         'event on publication {publication_id} not stored!',
        #         session=session._session_id,
        #         authid=session._authid,
        #         authrole=session._authrole,
        #         realm=session._realm,
        #         publication_id=publication_id)

        pub = cfxdb.realmstore.Publication()

        pub.timestamp = time_ns()
        pub.publication = publication_id
        pub.publisher = session._session_id

        pub.topic = publish.topic

        # FIXME: runs into pmap assert
        pub.args = list(publish.args) if type(publish.args) == tuple else publish.args

        pub.kwargs = publish.kwargs
        pub.payload = publish.payload
        pub.acknowledge = publish.acknowledge
        pub.retain = publish.retain
        pub.exclude_me = publish.exclude_me
        pub.exclude = publish.exclude
        pub.exclude_authid = publish.exclude_authid
        pub.exclude_authrole = publish.exclude_authrole
        pub.eligible = publish.eligible
        pub.eligible_authid = publish.eligible_authid
        pub.eligible_authrole = publish.eligible_authrole
        pub.enc_algo = publish.enc_algo
        pub.enc_key = publish.enc_key
        pub.enc_serializer = publish.enc_serializer

        self._schema.publications[txn, publication_id] = pub

    def store_event_history(self, publication_id: int, subscription_id: int, receiver: ISession):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.store_event_history`
        """
        assert type(publication_id) == int
        assert type(subscription_id) == int

        # FIXME: unexpected type <class 'backend.BackendSession'> for receiver
        # assert isinstance(receiver, RouterSession), 'unexpected type {} for receiver'.format(type(receiver))

        self._buffer.append([self._store_event_history, publication_id, subscription_id, receiver])

    def _store_event_history(self, txn, publication_id, subscription_id, receiver):

        # FIXME
        pub = self._schema.publications[txn, publication_id]
        if not pub:
            self.log.debug('no publication {publication_id} in store (schema.publications)',
                           publication_id=publication_id)
            return

        # FIXME
        # sub = self._schema.subscriptions[txn, subscription_id]
        # if not sub:
        #     self.log.warn('no subscription_id {subscription_id} in store (schema.subscriptions)', subscription_id=subscription_id)
        #     return

        receiver_session_id = receiver._session_id
        if not receiver_session_id:
            self.log.warn('no session ID for receiver (anymore) - will not store event!')
            return

        evt = cfxdb.realmstore.Event()
        evt.timestamp = time_ns()
        evt.subscription = subscription_id
        evt.publication = publication_id
        evt.receiver = receiver_session_id

        # FIXME
        # evt.retained = None
        # evt.acknowledged_delivery = None

        evt_key = (evt.subscription, np.datetime64(evt.timestamp, 'ns'))

        self._schema.events[txn, evt_key] = evt

    def get_events(self, subscription_id: int, limit: Optional[int] = None):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_events`
        """
        assert type(subscription_id) == int
        assert limit is None or type(limit) == int

        return self.get_event_history(subscription_id, from_ts=0, until_ts=time_ns(), reverse=True, limit=limit)

    def get_event_history(self,
                          subscription_id: int,
                          from_ts: int,
                          until_ts: int,
                          reverse: Optional[bool] = None,
                          limit: Optional[int] = None) -> Optional[List[Dict[str, Any]]]:
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_event_history`
        """
        assert type(subscription_id) == int
        assert type(from_ts) == int
        assert type(until_ts) == int
        assert type(reverse) == bool
        assert limit is None or type(limit) == int

        # FIXME
        # from_key = (subscription_id, np.datetime64(from_ts, 'ns'))
        from_key = (subscription_id, from_ts)
        # to_key = (subscription_id, np.datetime64(until_ts, 'ns'))
        to_key = (subscription_id, until_ts)

        with self._db.begin() as txn:
            res = []
            i = 0

            for evt in self._schema.events.select(txn,
                                                  from_key=from_key,
                                                  to_key=to_key,
                                                  return_keys=False,
                                                  reverse=True,
                                                  limit=limit):

                pub: Publication = self._schema.publications.select(txn, evt.publication)
                if pub:
                    res.append(pub.marshal())
                    i += 1
                    if limit is not None and i >= limit:
                        break

        return res

    def maybe_queue_call(self, session, call, registration, authorization):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.maybe_queue_call`
        """
        # FIXME: match this against the config, not just plain accept queueing!
        if registration.id not in self._queued_calls:
            self._queued_calls[registration.id] = deque()

        self._queued_calls[registration.id].append(QueuedCall(session, call, registration, authorization))

        return True

    def get_queued_call(self, registration):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_queued_call`
        """
        if registration.id in self._queued_calls and self._queued_calls[registration.id]:
            return self._queued_calls[registration.id][0]

    def pop_queued_call(self, registration):
        """
        Implements :meth:`crossbar._interfaces.IRealmStore.get_event_history`
        """
        if registration.id in self._queued_calls and self._queued_calls[registration.id]:
            return self._queued_calls[registration.id].popleft()


IRealmStore.register(RealmStoreDatabase)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import re
import psutil

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.utils import getProcessOutput

from autobahn.util import utcnow

from crossbar.edge.worker.monitor._base import Monitor

__all__ = ('CPUMonitor', )


class CPUMonitor(Monitor):
    """
    CPU load, frequency and temperature monitoring.
    """

    ID = u'cpu'

    # Physical id 3
    _PATH_SENSORS_PHYS_SOCKET_ID = re.compile(r"^Physical id (\d+)$")

    # Core 10
    _PATH_SENSORS_PHYS_CORE_ID = re.compile(r"^Core (\d+)$")

    def __init__(self, config=None):
        Monitor.__init__(self, config)

        self._cpu_last = None

        self._processors = {}
        self._sockets = []
        self._physid_to_id = {}
        self._id_to_physid = {}

        sockets = {}

        with open('/proc/cpuinfo', 'r') as fd:

            processor_id = None
            physical_socket_id = None
            physical_core_id = None

            for line in fd.readlines():

                line = line.strip()

                if line == "":
                    self._processors[processor_id] = (physical_socket_id, physical_core_id)
                    if physical_socket_id not in sockets:
                        sockets[physical_socket_id] = []
                    sockets[physical_socket_id].append(physical_core_id)
                else:
                    key, value = line.split(':')
                    key = key.strip()
                    value = value.strip()

                    if key == "processor":
                        processor_id = int(value)

                    elif key == "physical id":
                        physical_socket_id = int(value)

                    elif key == "core id":
                        physical_core_id = int(value)

        i = 0
        for pi in sorted(sockets.keys()):
            cores = []
            j = 0
            for pj in sorted(sockets[i]):
                cores.append(pj)
                self._physid_to_id[(pi, pj)] = (i, j)
                self._id_to_physid[(i, j)] = (pi, pj)
                j += 1
            self._sockets.append((pi, cores))
            i += 1

    @inlineCallbacks
    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        Monitor.poll(self)

        # create new, empty event
        #
        current = {
            # the UTC timestamp when measurement was taken
            u'timestamp': utcnow(),

            # the effective last period in secods
            u'last_period': self._last_period,

            # readings per CPU socket
            u'sockets': []
        }

        # fill in per-socket/per-core structures
        #
        for i in range(len(self._sockets)):
            socket = {
                # sequentially numbered socket ID
                u'id': i,

                # physical socket ID
                u'physical_id': self._sockets[i][0],

                # CPU socket temperature
                u'temperature': None,

                # CPU cores on this socket
                u'cores': []
            }
            for j in range(len(self._sockets[i][1])):
                core = {
                    # sequentially numbered core ID
                    u'id': j,

                    # physical core ID on this socket
                    u'physical_id': self._sockets[i][1][j],

                    # CPU core load
                    u'user': None,
                    u'system': None,
                    u'nice': None,
                    u'idle': None,
                    u'iowait': None,
                    u'irq': None,
                    u'softirq': None,
                    u'steal': None,
                    u'total': None,

                    # CPU core frequency
                    u'frequency': None,

                    # CPU core temperature
                    u'temperature': None
                }
                socket[u'cores'].append(core)
            current[u'sockets'].append(socket)

        # get current CPU load (via psutil)
        #
        cpu_now = psutil.cpu_times(percpu=True)

        if not self._cpu_last:
            self._cpu_last = cpu_now
        else:
            for i in range(len(cpu_now)):

                socket_id, core_id = self._physid_to_id[self._processors[i]]

                core = current[u'sockets'][socket_id][u'cores'][core_id]

                digits = 8

                # CPU core load stats
                core[u'user'] = round(cpu_now[i].user - self._cpu_last[i].user, digits)
                core[u'system'] = round(cpu_now[i].system - self._cpu_last[i].system, digits)
                core[u'nice'] = round(cpu_now[i].nice - self._cpu_last[i].nice, digits)
                core[u'idle'] = round(cpu_now[i].idle - self._cpu_last[i].idle, digits)
                core[u'iowait'] = round(cpu_now[i].iowait - self._cpu_last[i].iowait, digits)
                core[u'irq'] = round(cpu_now[i].irq - self._cpu_last[i].irq, digits)
                core[u'softirq'] = round(cpu_now[i].softirq - self._cpu_last[i].softirq, digits)
                core[u'steal'] = round(cpu_now[i].steal - self._cpu_last[i].steal, digits)

                # total CPU core load (= user + nice + system)
                core[u'total'] = round((cpu_now[i].user + cpu_now[i].nice + cpu_now[i].system) -
                                       (self._cpu_last[i].user + self._cpu_last[i].nice + self._cpu_last[i].system),
                                       digits)

                # normalize with effective period
                diff = self._last_period or 1.
                for k in core:
                    if (k != 'id' and k != 'physical_id' and k != 'frequency' and k != 'temperature'):
                        core[k] = round(core[k] / diff, digits)

            self._cpu_last = cpu_now

        # get current CPU frequency (from procfs)
        #
        with open('/proc/cpuinfo', 'r') as fd:

            physical_socket_id = None
            physical_core_id = None
            frequency = None

            for line in fd.readlines():

                line = line.strip()

                if line == "":
                    socket_id, core_id = self._physid_to_id[(physical_socket_id, physical_core_id)]
                    core = current[u'sockets'][socket_id][u'cores'][core_id]
                    core[u'frequency'] = frequency
                else:
                    key, value = line.split(':')
                    key = key.strip()
                    value = value.strip()

                    if key == "physical id":
                        physical_socket_id = int(value)

                    elif key == "core id":
                        physical_core_id = int(value)

                    elif key == "cpu MHz":
                        frequency = float(value)

        # get current CPU temperature (via /usr/bin/sensors)
        #
        res = yield getProcessOutput('/usr/bin/sensors')

        physical_socket_id = None
        physical_core_id = None
        # socket_temperature = None
        core_temperature = -1

        for line in res.splitlines():

            line = line.strip()

            if line == "":
                pass
            else:
                if line.startswith("Physical"):
                    key, value = line.split(':')
                    match = self._PATH_SENSORS_PHYS_SOCKET_ID.match(key)
                    if match:
                        physical_socket_id = int(match.groups()[0])
                        value = value.strip()
                        # socket_temperature = float(value[1:5])

                elif line.startswith("Core"):
                    key, value = line.split(':')
                    match = self._PATH_SENSORS_PHYS_CORE_ID.match(key)
                    if match:
                        physical_core_id = int(match.groups()[0])
                        value = value.strip()
                        core_temperature = float(value[1:5])

                        socket_id, core_id = self._physid_to_id[(physical_socket_id, physical_core_id)]

                        core = current[u'sockets'][socket_id][u'cores'][core_id]
                        core[u'temperature'] = core_temperature

        self._last_value = current

        returnValue(self._last_value)



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from crossbar.master.api.remote import RemoteApi

__all__ = ('RemoteWorkerApi', )


class RemoteWorkerApi(RemoteApi):

    PREFIX = u'crossbarfabriccenter.remote.worker.'

    PROCS = {
        # these are worker level procedures
        u'worker': [
            u'shutdown',
            u'get_status',
            u'get_pythonpath',
            u'add_pythonpath',
            u'get_cpu_affinity',
            u'set_cpu_affinity',
            u'get_profilers',
            u'start_profiler',
            u'get_profile',
            u'get_process_info',
            u'get_process_stats',
            u'set_process_stats_monitoring',
        ],
    }

    EVENTS = {
        # these are worker level topics
        u'worker': [
            u'on_worker_log',
            u'on_profile_started',
            u'on_profile_finished',
        ]
    }
